from iam_syncr.errors import InvalidDocument, BadPolicy, ProgrammerError
from iam_syncr.helpers import listify, listified, as_list

import logging
import json
import six

log = logging.getLogger("iam_syncr.roles")

class Statements(object):
    def __init__(self, name, self_type, account_id, accounts, location=None):
        self.name = name
        self.location = location
        self.self_type = self_type
        if self_type not in ("role", "bucket", "key"):
            raise ProgrammerError("Statements can only be instantiated with self_type of role or bucket or key\tgot={0}".format(self_type))

        self.accounts = accounts
        self.account_id = account_id

    def expand_trust_statement(self, statement, allow=False):
        """Make a trust statement"""
        result = dict((key, val) for key, val in statement.items() if key[0].isupper())
        principal = "principal" if allow else "notprincipal"
        self.expand_principal({principal: statement}, result)

        if "Action" not in result:
            result["Action"] = "sts:AssumeRole"

        if "Effect" not in result:
            result["Effect"] = "Allow"

        if "Sid" not in result:
            result["Sid"] = ""

        yield result

    def expand_principal(self, statement, result):
        """Expand out Principal and NotPrincipal"""
        for principal, capd in [("principal", "Principal"), ("notprincipal", "NotPrincipal")]:
            if principal in statement:
                if capd not in result:
                    result[capd] = {}

                looking_at = statement[principal]
                if not isinstance(looking_at, list):
                    looking_at = [looking_at]

                for instruction in looking_at:
                    principal = result[capd]

                    for specified in listified(instruction, "service"):
                        if specified == "ec2":
                            specified = "ec2.amazonaws.com"
                        listify(principal, "Service").append(specified)

                    for specified in listified(instruction, "federated"):
                        listify(principal, "Federated").extend(self.iam_arns_from_specification(specified))
                        if "Action" not in result:
                            result["Action"] = "sts:AssumeRoleWithSAML"

                    if "iam" in instruction:
                        listify(principal, "AWS").extend(self.iam_arns_from_specification(instruction))

        # Amazon gets rid of the lists if only one item
        # And this mucks around with the diffing....
        for princ in (result.get("Principal"), result.get("NotPrincipal")):
            if princ:
                for principal_type in ("AWS", "Federated", "Service"):
                    if principal_type in princ:
                        if len(listify(princ, principal_type)) == 1:
                            princ[principal_type] = princ[principal_type][0]
                        else:
                            princ[principal_type] = sorted(princ[principal_type])

    def make_permission_statements(self, policy, allow=None):
        """
        Make zero or more permissions statements from a policy definition

        If allow is None and no allow or Effect is specified, an error is raised
        Otherwise we assume the truthiness of allow
        """
        result = dict((key, val) for key, val in policy.items() if key[0].isupper())

        if allow is None and "Effect" not in policy:
            if not isinstance(policy.get("allow"), bool):
                raise BadPolicy("Need to specify whether we allow this policy or not", policy=policy, **{self.self_type:self.name})
            result["Effect"] = ["Deny", "Allow"][policy["allow"]]
        elif allow is not None:
            result["Effect"] = ["Deny", "Allow"][allow]

        for key, dest in (("action", "Action"), ("notaction", "NotAction")):
            for specified in listified(policy, key):
                listify(result, dest).append(specified)

        for key, dest in (("resource", "Resource"), ("notresource", "NotResource")):
            for specified in listified(policy, key):
                listify(result, dest).extend(self.fill_out_resources(specified))

        self.expand_principal(policy, result)

        if "Resource" not in result and "NotResource" not in result:
            raise BadPolicy("No Resource or NotResource was defined for policy", policy=policy, **{self.self_type:self.name})

        if "Action" not in result and "NotAction" not in result:
            raise BadPolicy("No Action or NotAction defined for this policy", policy=policy, **{self.self_type:self.name})

        if "Sid" not in result and self.self_type == "bucket":
            result["Sid"] = ""

        # Amazon gets rid of the lists if only one item
        # And this mucks around with the diffing....
        for thing in ("Action", "NotAction", "Resource", "NotResource"):
            if thing in result:
                if len(listify(result, thing)) == 1:
                    result[thing] = result[thing][0]
                else:
                    result[thing] = sorted(result[thing])

        yield result

    def fill_out_resources(self, resources):
        """Fill out the resources"""
        for resource in as_list(resources):
            if isinstance(resource, six.string_types):
                yield resource
            elif isinstance(resource, dict):
                for resource in self.expand_resource(resource):
                    yield resource
            else:
                raise BadPolicy("Resource should be a string or a dictionary", resource=resource)

    def expand_resource(self, resource):
        """Return a resource string for given resource"""
        if "iam" in resource:
            for found in self.iam_arns_from_specification(resource):
                yield found

        elif "kms" in resource:
            for found in self.kms_arns_from_specification(resource):
                yield found

        elif "sns" in resource:
            for found in self.sns_arns_from_specification(resource):
                yield found

        elif "s3" in resource:
            for bucket_key in listify(resource, "s3"):
                if bucket_key == "__self__":
                    if self.self_type == "role":
                        raise BadPolicy("Role policy has no __self__ bucket", role=self.name)
                    elif self.self_type == "key":
                        raise BadPolicy("Key policy has no __self__ bucket", key=self.name)
                    else:
                        bucket_key = self.name
                yield "arn:aws:s3:::{0}".format(bucket_key)
                if '/' not in bucket_key:
                    yield "arn:aws:s3:::{0}/*".format(bucket_key)

        else:
            raise BadPolicy("Unknown resource type", resource=resource)

    def sns_arns_from_specification(self, resource):
        """Get us kms arns from this specification"""
        for key_id in listify(resource, "sns"):
            location = self.location
            if 'location' in resource:
                location = resource['location']

            provided_accounts = resource.get("account", "")
            if not isinstance(provided_accounts, list):
                provided_accounts = [provided_accounts]

            for provided_account in provided_accounts:
                if provided_account:
                    if provided_account not in self.accounts:
                        raise BadPolicy("Unknown account specified", account=provided_account, specification=resource)
                    else:
                        account_id = self.accounts[provided_account]
                else:
                    account_id = self.account_id

                yield "arn:aws:sns:{0}:{1}:{2}".format(location, account_id, key_id)


    def kms_arns_from_specification(self, resource):
        """Get us kms arns from this specification"""
        for key_id in listify(resource, "kms"):
            alias = None
            if key_id == "__self__":
                if self.self_type != "key":
                    raise BadPolicy("No __self__ key for this policy", type=self.self_type, name=self.name)
                else:
                    alias = self.name
                    location = self.location

            if not alias:
                if isinstance(key_id, six.string_types):
                    alias = key_id
                    location = resource.get("location", self.location)
                else:
                    alias = key_id.get("alias")
                    location = key_id.get("location", resource.get("location", self.location))
                    key_id = key_id.get("key_id")

            provided_accounts = resource.get("account", "")
            if not isinstance(provided_accounts, list):
                provided_accounts = [provided_accounts]

            for provided_account in provided_accounts:
                if provided_account:
                    if provided_account not in self.accounts:
                        raise BadPolicy("Unknown account specified", account=provided_account, specification=resource)
                    else:
                        account_id = self.accounts[provided_account]
                else:
                    account_id = self.account_id

                if alias:
                    yield "arn:aws:kms:{0}:{1}:alias/{2}".format(location, account_id, alias)
                else:
                    yield "arn:aws:kms:{0}:{1}:key/{2}".format(location, account_id, key_id)

    def iam_arns_from_specification(self, specification):
        """Get us an iam arn from this specification"""
        if not isinstance(specification, list):
            specification = [specification]

        for spec in specification:
            if isinstance(spec, six.string_types):
                yield spec
            else:
                provided_accounts = spec.get("account", "")
                if not isinstance(provided_accounts, list):
                    provided_accounts = [provided_accounts]

                for provided_account in provided_accounts:
                    if provided_account:
                        if provided_account not in self.accounts:
                            raise BadPolicy("Unknown account specified", account=provided_account, specification=spec)
                        else:
                            account_id = self.accounts[provided_account]
                    else:
                        account_id = self.account_id

                    users = spec.get("users", [])
                    for name in listified(spec, "iam"):
                        if name == "__self__":
                            if self.self_type == 'bucket':
                                raise BadPolicy("Bucket policy has no __self__ iam role", bucket=self.name)

                            account_id = self.account_id
                            name = "role/{0}".format(self.name)

                        service = "sts" if name.startswith("assumed-role") else "iam"
                        arn = "arn:aws:{0}::{1}:{2}".format(service, account_id, name)
                        if not users:
                            yield arn
                        else:
                            for user in users:
                                yield "{0}/{1}".format(arn, user)

    def make_document(self, statements):
        """Make sure our document is valid and return it formatted correctly"""
        if not isinstance(statements, list):
            raise Exception("Statements should be a list!: got {0}".format(statements))

        document = {"Version": "2012-10-17", "Statement": statements}

        try:
            return json.dumps(document, indent=2)
        except (TypeError, ValueError) as err:
            raise InvalidDocument("Document wasn't valid json", error=err, **{self.self_type:self.name})




from iam_syncr.errors import SyncrError, BadConfiguration, InvalidConfiguration, NoConfiguration
from iam_syncr.amazon.base import Amazon
from iam_syncr.syncer import Sync
from iam_syncr import VERSION

from rainbow_logging_handler import RainbowLoggingHandler
import argparse
import logging
import fnmatch
import yaml
import sys
import os

log = logging.getLogger("iam_sync.executor")

def setup_logging(verbose=False):
    log = logging.getLogger("")
    handler = RainbowLoggingHandler(sys.stderr)
    handler._column_color['%(asctime)s'] = ('cyan', None, False)
    handler._column_color['%(levelname)-7s'] = ('green', None, False)
    handler._column_color['%(message)s'][logging.INFO] = ('blue', None, False)
    handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)-7s %(name)-15s %(message)s"))
    log.addHandler(handler)
    log.setLevel([logging.INFO, logging.DEBUG][verbose])

    logging.getLogger("boto").level = logging.CRITICAL

def argparse_readable_folder(value):
    """Argparse type for a readable folder"""
    if not os.path.exists(value):
        raise argparse.ArgumentTypeError("{0} doesn't exist".format(value))
    if not os.path.isdir(value):
        raise argparse.ArgumentTypeError("{0} exists but isn't a folder".format(value))
    if not os.access(value, os.R_OK):
        raise argparse.ArgumentTypeError("{0} exists and is a folder but isn't readable".format(value))
    return os.path.abspath(value)

def make_parser():
    """Make us a parser"""
    parser = argparse.ArgumentParser(description="Sync script, supply your own creds!")
    parser.add_argument("-v", "--verbose"
        , help = "Show debug logging"
        , action = "store_true"
        )

    parser.add_argument("folder"
        , help = "The folder containing the roles we want to sync"
        , type = argparse_readable_folder
        )

    parser.add_argument("--accounts-location"
        , help = "Path to accounts.yaml holding the map of human names to accounts ids"
        )

    parser.add_argument("--filename-match"
        , help = "A glob to match the path of the configuration against (relative to the specified folder)"
        , default = "*.yaml"
        )

    parser.add_argument("--only-consider"
        , help = "Only sync these (i.e. roles, remove_roles, users)"
        , action = "append"
        )

    parser.add_argument("--dry-run"
        , help = "Print out what policies would be set/removed"
        , action = "store_true"
        )

    return parser

def accounts_from(location):
    """Get the accounts dictionary"""
    if not os.path.exists(location):
        raise SyncrError("Could not find an accounts.yaml", location=location)

    if not os.access(location, os.R_OK):
        raise SyncrError("Could not read the accounts.yaml", location=location)

    try:
        accounts = yaml.load(open(location))
    except yaml.parser.ParserError as error:
        raise SyncrError("Failed to parse the accounts yaml file", location=location, error_typ=error.__class__.__name__, error=error)

    for account_id in list(accounts.values()):
        if account_id not in accounts:
            accounts[account_id] = account_id

    return accounts

def make_amazon(folder, accounts_location=None, dry_run=False):
    """Find the account we're using and return a setup Amazon object"""
    if not accounts_location:
        accounts_location = os.path.join(folder, '..', 'accounts.yaml')

    accounts = accounts_from(accounts_location)
    account_name = os.path.basename(folder)

    if account_name not in accounts:
        raise SyncrError("Please add this account to accounts.yaml", accounts_yaml_location=accounts_location, account_name=account_name)
    account_id = accounts[account_name]

    amazon = Amazon(account_id, account_name, accounts, dry_run=dry_run)
    amazon.setup()

    return amazon

def do_sync(amazon, found, only_consider=None):
    """Sync the configuration from this folder"""
    try:
        parsed = parse_configurations(found)
    except BadConfiguration as err:
        log.error("Failed to parse all the yaml specifications")
        for _, error in sorted(err.kwargs["parse_errors"].items()):
            log.error(error)
        raise BadConfiguration()

    sync = Sync(amazon)
    sync.register_default_types()

    if only_consider:
        dont_consider = [considering for considering in only_consider if considering not in sync.types]
        if dont_consider:
            raise SyncrError("Told to sync unknown types", only_sync=list(sync.types.keys()), unknown_types=dont_consider)

    for location, configuration in sorted(parsed.items()):
        sync.add(configuration, location, only_consider)

    try:
        log.info("Combining configuration")
        combined = sync.combine_configurations()
    except BadConfiguration as err:
        log.error("Your configuration didn't make sense")
        for error in err.kwargs["errors"]:
            log.error(error)
        raise BadConfiguration()

    log.info("Starting sync")
    sync.sync(combined)

def parse_configurations(locations):
    """
    Return a dictionary of {location: <parsed_yaml>} for .yaml files in this folder

    Or Raise A BadConfiguration(parse_errors={<location>: <parse_error>})
    With all the errors that are encountered
    """
    parsed = {}
    parse_errors = {}
    for location in locations:
        try:
            config = yaml.load(open(location))
            if not isinstance(config, dict):
                parse_errors[location] = InvalidConfiguration("Configuration is not a dictionary", location=location, found=type(config))
            else:
                parsed[location] = config
        except yaml.parser.ParserError as err:
            parse_errors[location] = InvalidConfiguration("Couldn't parse the yaml", location=location, err_type=err.__class__.__name__, err=err)

    if parse_errors:
        raise BadConfiguration(parse_errors=parse_errors)

    return parsed

def find_configurations(folder, filename_match):
    """Find all the configurations in this folder"""
    found = []
    for root, dirs, files in os.walk(folder):
        for filename in files:
            location = os.path.join(root, filename)
            relative_location = os.path.relpath(location, start=folder)
            if fnmatch.fnmatch(relative_location, filename_match):
                found.append(location)

    if not found:
        raise NoConfiguration(folder=folder)

    return found

def main(argv=None):
    parser = make_parser()
    args = parser.parse_args(argv)
    setup_logging(verbose=args.verbose)

    Amazon.set_boto_useragent("iam_syncr", VERSION)

    try:
        log.info("Making a connection to amazon")
        amazon = make_amazon(folder=args.folder, accounts_location=args.accounts_location, dry_run=args.dry_run)

        log.info("Finding the configuration")
        found = find_configurations(args.folder, args.filename_match)

        log.info("Syncing for account %s from %s", amazon.account_id, args.folder)
        do_sync(amazon, found, args.only_consider)

        if not amazon.changes:
            log.info("No changes were made!")
    except SyncrError as err:
        print("!" * 80)
        print("Something went wrong => {0} |:| {1}".format(err.__class__.__name__, err))
        sys.exit(1)

if __name__ == '__main__':
    main()




from iam_syncr.statements import Statements
from iam_syncr.amazon.kms import AmazonKms
from iam_syncr.helpers import listified
from iam_syncr.errors import BadPolicy

import logging

log = logging.getLogger("iam_syncr.kms")

class Kms(object):
    def __init__(self, name, definition, amazon, templates=None):
        self.name = name
        self.grant = []
        self.amazon = amazon
        self.definition = definition
        self.statements = Statements(name, "key", amazon.account_id, amazon.accounts, location=self.definition.get("location"))
        self.permission = list(self.statements.make_permission_statements({"principal": {"iam": "root"}, "action": "kms:*", "resource": "*", "Sid": ""}, allow=True))

    def setup(self):
        """Raise errors if the definition doesn't make sense"""
        self.location = self.definition.get("location")
        self.description = self.definition.get("description")

        self.connection = self.amazon.kms_connection_for(self.location)
        if not self.description:
            raise BadPolicy("Please define a description", key=self.name)

        if "admin_users" in self.definition:
            policy = {"principal": self.definition["admin_users"], "action": "kms:*", "resource": { "kms": "__self__" }, "Sid": ""}
            for statement in self.statements.make_permission_statements(policy, allow=True):
                self.permission.append(statement)

        for key, default_allow in (("permission", None), ("allow_permission", True), ("deny_permission", False)):
            for policy in listified(self.definition, key):
                for statement in self.statements.make_permission_statements(policy, allow=default_allow):
                    self.permission.append(statement)

        for policy in listified(self.definition, "grant"):
            if not isinstance(policy, dict):
                raise BadPolicy("Grant must be a dictionary of atleast {grantee, operations}", got=policy, key=self.name)

            if "grantee" not in policy and "operations" not in policy:
                raise BadPolicy("Grant must be a dictionary of atleast {grantee, operations}", got=list(policy.keys()), key=self.name)

            policy['grantee'] = list(self.statements.iam_arns_from_specification(policy['grantee']))
            if 'retiree' in policy:
                policy['retiree'] = list(self.statements.iam_arns_from_specification(policy['retiree']))

            for thing in ('grantee', 'retiree'):
                if thing in policy:
                    if len(policy[thing]) > 1:
                        raise BadPolicy("Can only have one {0}".format(thing), got=len(policy[thing]), key=self.name)
                    policy[thing] = policy[thing][0]

            self.grant.append(policy)

    def resolve(self):
        """Make sure this key exists and has only what policies we want it to have"""
        permission_document = self.make_permission_document(self.permission)

        amazon_keys = AmazonKms(self.amazon, self.connection)
        key_info = amazon_keys.key_info(self.name)
        if not key_info:
            amazon_keys.create_key(self.name, self.description, permission_document=permission_document)
        else:
            amazon_keys.modify_key(self.name, self.description, permission_document=permission_document)

        amazon_keys.modify_grant(self.name, self.description, grant=self.grant)

    def make_permission_document(self, permissions):
        """Return a document for these permissions, or None if no permissiosn"""
        if not permissions:
            return
        return self.statements.make_document(permissions)




from iam_syncr.errors import SyncrError

from boto.iam.connection import IAMConnection
from boto.s3.connection import S3Connection
import logging
import boto
import sys
import six

if six.PY2:
    KMSConnection = None
else:
    from boto.kms.layer1 import KMSConnection

log = logging.getLogger("iam_syncr.amazon.base")

class Amazon(object):
    dry_run = False
    connection = None

    # Account info that is overridden by __init__
    accounts = None
    account_id = None
    account_name = None

    @classmethod
    def set_boto_useragent(self, app_name, version):
        """Put this app in the useragent used by boto"""
        __import__("boto")
        useragent = sys.modules["boto.connection"].UserAgent
        if app_name not in useragent:
            sys.modules["boto.connection"].UserAgent = "{0} {1}/{2}".format(useragent, app_name, version)

    def __init__(self, account_id, account_name, accounts, dry_run=False):
        self.changes = False
        self.dry_run = dry_run
        self.accounts = accounts
        self.account_id = account_id
        self.account_name = account_name

    def setup(self):
        """Make sure our current credentials are for this account and set self.connection"""
        try:
            connection = IAMConnection()
        except boto.exception.NoAuthHandlerFound:
            raise SyncrError("Export AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY before running this script (your aws credentials)")

        try:
            result = connection.list_roles()
        except boto.exception.BotoServerError as error:
            if error.status == 403:
                raise SyncrError("Your credentials aren't allowed to look at iam :(")
            else:
                raise

        roles = self.all_roles = result["list_roles_response"]["list_roles_result"]["roles"]
        if not roles:
            raise SyncrError("There are no roles in your account, I can't figure out the account id")

        amazon_account_id = roles[0]['arn'].split(":")[4]
        if str(self.account_id) != str(amazon_account_id):
            raise SyncrError("Please use credentials for the right account", expect=self.account_id, got=amazon_account_id)

        # If reached this far, the credentials belong to the correct account :)
        self.connection = connection
        return connection

    @property
    def s3_connection(self):
        if getattr(self, "_s3_connection", None) is None:
            try:
                self._s3_connection = S3Connection()
            except boto.exception.NoAuthHandlerFound:
                raise SyncrError("Export AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY before running this script (your aws credentials)")
        return self._s3_connection

    def kms_connection_for(self, location):
        if KMSConnection is None:
            raise SyncrError("Sorry, need python3 to do anything related to kms")

        if getattr(self, "_kms_connections", None) is None:
            self._kms_connections = {}

        if location not in self._kms_connections:
            try:
                self._kms_connections[location] = boto.kms.connect_to_region(location)
            except boto.exception.NoAuthHandlerFound:
                raise SyncrError("Export AWS_ACCESS_KEY_ID and AWS_SECRET_ACCESS_KEY before running this script (your aws credentials)")
        return self._kms_connections[location]




from iam_syncr.errors import BadRole, CantFindTemplate, NoTemplates
from iam_syncr.amazon.roles import AmazonRoles
from iam_syncr.statements import Statements
from iam_syncr.helpers import listified

from option_merge import MergedOptions
import logging
import six

log = logging.getLogger("iam_syncr.roles")

class RoleRemoval(object):
    def __init__(self, name, amazon, templates=None):
        self.name = name
        self.amazon = amazon
        self.templates = templates

    def setup(self):
        """Make sure our name is a string"""
        if not isinstance(self.name, six.string_types):
            raise BadRole("Told to remove a role, but not specified as a string", name=self.name, found_type=type(self.name))

    def resolve(self):
        """Remove the role"""
        AmazonRoles(self.amazon).remove_role(self.name)

class Role(object):
    def __init__(self, name, definition, amazon, templates=None):
        self.name = name
        self.templates = templates
        self.definition = definition
        self.statements = Statements(name, "role", amazon.account_id, amazon.accounts)
        self.policy_name = "syncr_policy_{0}".format(self.name.replace("/", "__"))

        self.amazon = amazon
        self.amazon_roles = AmazonRoles(amazon)

        self.trust = []
        self.distrust = []
        self.permission = []

    def setup(self):
        """Raise errors if the definition doesn't make sense"""
        if "use" in self.definition:
            template = self.definition["use"]
            if not self.templates:
                raise NoTemplates(name=self.name, looking_for_template=template, available=self.templates.keys())

            if template not in self.templates:
                raise CantFindTemplate(name=self.name, looking_for_template=template, available=self.templates.keys())

            self.definition = MergedOptions.using(self.templates[template], self.definition)

        self.description = self.definition.get("description", "No description provided!")

        for statement in listified(self.definition, "allow_to_assume_me"):
            self.trust.extend(self.statements.expand_trust_statement(statement, allow=True))

        for statement in listified(self.definition, "disallow_to_assume_me"):
            self.distrust.extend(self.statements.expand_trust_statement(statement, allow=False))

        for key, default_allow in (("permission", None), ("allow_permission", True), ("deny_permission", False)):
            for policy in listified(self.definition, key):
                for statement in self.statements.make_permission_statements(policy, allow=default_allow):
                    self.permission.append(statement)

    def resolve(self):
        """Make sure this user exists and has only what policies we want it to have"""
        # Get the permission and trust document
        # Make sure they're both valid before continuing
        trust_document = self.make_trust_document(self.trust, self.distrust)
        permission_document = self.make_permission_document(self.permission)

        role_info = self.amazon_roles.role_info(self.name)
        if not role_info:
            self.amazon_roles.create_role(self.name, trust_document, policies={self.policy_name: permission_document})
        else:
            self.amazon_roles.modify_role(role_info, self.name, trust_document, policies={self.policy_name: permission_document})

        if self.definition.get("make_instance_profile"):
            self.amazon_roles.make_instance_profile(self.name)

    def make_trust_document(self, trust, distrust):
        """Make a document for trust or None if no trust or distrust"""
        if not trust and not distrust:
            return

        return self.statements.make_document((trust or []) + (distrust or []))

    def make_permission_document(self, permissions):
        """Return a document for these permissions, or None if no permissiosn"""
        if not permissions:
            return
        return self.statements.make_document(permissions)




from iam_syncr.amazon.documents import AmazonDocuments
from iam_syncr.amazon.common import AmazonMixin
from iam_syncr.errors import BadPolicy

from boto.s3.tagging import TagSet, Tags
from itertools import chain
import logging
import boto

log = logging.getLogger("iam_syncr.amazon.buckets")

class AmazonBuckets(AmazonMixin, object):
    def __init__(self, amazon):
        self.amazon = amazon
        self.documents = AmazonDocuments()
        self.connection = amazon.s3_connection

    def bucket_info(self, name):
        """Return what amazon knows about this bucket"""
        try:
            return self.connection.get_bucket(name)
        except boto.exception.S3ResponseError as error:
            if error.status == 404:
                return False
            raise

    def current_policy(self, bucket):
        """Return the current policy for this bucket"""
        try:
            return bucket.get_policy().decode('utf-8')
        except boto.exception.S3ResponseError as error:
            if error.status == 404:
                return "{}"
            raise

    def current_tags(self, bucket):
        """Return the tags associated with this bucket"""
        try:
            return dict(chain.from_iterable([(tag.key, tag.value) for tag in tags] for tags in bucket.get_tags()))
        except boto.exception.S3ResponseError as error:
            if error.status == 404:
                return {}
            raise

    def has_bucket(self, name):
        """Return whether amazon has info about this role"""
        return bool(self.bucket_info(name))

    def create_bucket(self, name, location, permission_document=None, tags=None):
        """Create a role"""
        with self.catch_boto_400("Couldn't create bucket", name=name):
            for _ in self.change("+", "bucket[{0}] ".format(location), name=name):
                self.connection.create_bucket(name, location=location)

        # And add our permissions
        if permission_document:
            with self.catch_boto_400("Couldn't add policy", "Bucket {0} - Permission document".format(name), permission_document, bucket=name):
                for _ in self.change("+", "bucket_policy", bucket=name, document=permission_document):
                    self.bucket_info(name).set_policy(permission_document)

    def modify_bucket(self, name, location, permission_document, tags):
        """Modify a bucket"""
        log.info("Inspecting bucket\tname=%s", name)
        bucket = self.bucket_info(name)
        if not bucket:
            return

        current_location = bucket.get_location()
        if current_location != location:
            raise BadPolicy("The location of the bucket is wrong. You need to delete and recreate the bucket to have it in your specified location", current=current_location, wanted=location)

        current_policy = self.current_policy(bucket)
        changes = list(self.documents.compare_two_documents(current_policy, permission_document))
        if changes:
            with self.catch_boto_400("Couldn't modify policy", "Bucket {0} policy".format(name), permission_document, bucket=name):
                for _ in self.change("M", "bucket_policy", bucket=name, changes=changes):
                    bucket.set_policy(permission_document)

        self.modify_bucket_tags(name, bucket, tags)

    def modify_bucket_tags(self, name, bucket, tags):
        """Modify the tags on a bucket"""
        changes = {}
        new_tags = TagSet()
        current_tags = self.current_tags(bucket)

        for tag_name, tag_val in tags.items():
            if tag_name in current_tags:
                if current_tags[tag_name] != tag_val:
                    changes[tag_name] = ("modify", tag_name, current_tags[tag_name], tag_val)
            elif tag_name not in current_tags:
                changes[tag_name] = ("create", tag_name, None, tag_val)

            new_tags.add_tag(tag_name, tag_val)

        for tag_name in current_tags:
            if tag_name not in tags:
                changes[tag_name] = ("delete", tag_name, current_tags[tag_name], None)

        if changes:
            if not new_tags:
                for _ in self.change("D", "bucket_tags", bucket=name, changes=["Delete all tags"]):
                    bucket.delete_tags()
            else:
                one_letter = "M" if any(typ in ("modify", "delete") for typ, _, _, _ in changes.values()) else "C"
                for _ in self.change(one_letter, "bucket_tag", bucket=name, changes=["{0} {1} from {2} to {3}".format(*change) for change in changes.values()]):
                    t = Tags()
                    t.add_tag_set(new_tags)
                    bucket.set_tags(t)




from iam_syncr.amazon.documents import AmazonDocuments
from iam_syncr.amazon.common import AmazonMixin
from iam_syncr.errors import BadAlias, BadRole

import logging
import boto
import six

if not six.PY2:
    import boto.kms.exceptions

log = logging.getLogger("iam_syncr.amazon.kms")

class AmazonKms(AmazonMixin, object):
    def __init__(self, amazon, connection):
        self.amazon = amazon
        self.documents = AmazonDocuments()
        self.connection = connection

    def key_info(self, alias):
        """Return what amazon knows about this key"""
        try:
            return self.connection.describe_key("alias/{0}".format(alias))["KeyMetadata"]
        except boto.kms.exceptions.NotFoundException:
            return False

    def current_policy(self, key):
        """Return the current policy for this key"""
        try:
            return self.connection.get_key_policy(key["KeyId"], "default")["Policy"]
        except boto.kms.exceptions.NotFoundException:
            return "{}"

    def has_key(self, alias):
        """Return whether amazon has info about this key"""
        return bool(self.key_info(alias))

    def create_key(self, alias, description, permission_document=None):
        """Create a key"""
        with self.catch_boto_400("Couldn't create key", document=permission_document, alias=alias):
            for _ in self.change("+", "key", alias=alias):
                key = self.connection.create_key(permission_document, description)["KeyMetadata"]
                self.connection.create_alias("alias/{0}".format(alias), key["KeyId"])

    def modify_key(self, alias, description, permission_document):
        """Modify a key"""
        key = self.key_info(alias)
        if not key:
            return

        current_description = key["Description"]
        if current_description != description:
            for _ in self.change("M", "key_description", key=alias, description=description):
                self.connection.update_key_description(key["KeyId"], description)

        current_policy = self.current_policy(key)
        changes = list(self.documents.compare_two_documents(current_policy, permission_document))
        if changes:
            with self.catch_boto_400("Couldn't modify policy", "Key {0} policy".format(alias), permission_document, key=alias):
                for _ in self.change("M", "key_policy", key=alias, changes=changes, description=description):
                    self.connection.put_key_policy(key["KeyId"], 'default', permission_document)

    def modify_grant(self, alias, description, grant):
        """Modify grants on a key"""
        key = self.key_info(alias)
        if not key:
            raise BadAlias("Where did the key go?", alias=alias)

        key_id = key["KeyId"]
        new_grants = []
        current_grants = self.connection.list_grants(key_id)["Grants"]
        for policy in current_grants:
            policy["Operations"] = sorted(policy["Operations"])

        for policy in grant:
            nxt = {"GranteePrincipal": self.user_from_arn(policy["grantee"]), "RetireePrincipal": self.user_from_arn(policy.get("Retiree")), "Operations": sorted(policy["operations"]), "Constraints": policy.get("constraints"), "GrantTokens": policy.get("grant_tokens")}
            nxt = dict((key, val) for key, val in nxt.items() if val is not None)

            if not any(all(current[key] == val for key, val in nxt.items()) for current in current_grants):
                new_grants.append(policy)

        for policy in new_grants:
            for _ in self.change("+", "key_grant", key=alias, grantee=policy["grantee"]):
                self.connection.create_grant(key_id, policy["grantee"], retiring_principal=policy.get("retiree"), operations=policy["operations"], constraints=policy.get("constraints"), grant_tokens=policy.get("grant_tokens"))

    def user_from_arn(self, arn):
        """Convert an arn into the user id"""
        if arn is None:
            return

        role_ids = [item["role_id"] for item in self.amazon.all_roles if item['arn'] == arn]
        if len(role_ids) != 1:
            raise BadRole("Didn't find a single role id for specified arn", got=role_ids, arn=arn)

        return role_ids[0]




from iam_syncr.amazon.common import AmazonMixin, LeaveAlone
from iam_syncr.amazon.documents import AmazonDocuments

from six.moves.urllib import parse
import logging
import boto
import json

log = logging.getLogger("iam_syncr.amazon.roles")

class AmazonRoles(AmazonMixin, object):
    def __init__(self, amazon):
        self.amazon = amazon
        self.documents = AmazonDocuments()
        self.connection = amazon.connection

    def split_role_name(self, name):
        """Split a role name into it's (name, path)"""
        split = name.split('/')
        role_name = split[-1]
        path = '/'.join(split[0:-1]) or None
        if path and not path.startswith('/'):
            path = "/{0}".format(path)
        if path and not path.endswith("/"):
            path = "{0}/".format(path)
        return role_name, path

    def role_info(self, name):
        """Return what amazon knows about this role"""
        try:
            role_name, _ = self.split_role_name(name)
            return self.connection.get_role(role_name)["get_role_response"]["get_role_result"]
        except boto.exception.BotoServerError as error:
            if error.status == 404:
                return False
            raise

    def has_role(self, name):
        """Return whether amazon has info about this role"""
        return bool(self.role_info(name))

    def info_for_profile(self, name):
        """Return what roles are attached to this profile if it exists"""
        profiles = []
        role_name, _ = self.split_role_name(name)
        with self.ignore_boto_404():
            with self.catch_boto_400("Couldn't list instance profiles associated with a role", role=role_name):
                result = self.connection.list_instance_profiles_for_role(role_name)
            profiles = result["list_instance_profiles_for_role_response"]["list_instance_profiles_for_role_result"]["instance_profiles"]

        existing_profile = None
        for profile in profiles:
            if profile["instance_profile_name"] == name:
                existing_profile = profile
                break

        if existing_profile:
            if "member" in existing_profile["roles"]:
                return [existing_profile["roles"]["member"]["role_name"]]
            else:
                return []

    def make_instance_profile(self, name):
        """Make an instance profile with this name containing this role"""
        role_name, _ = self.split_role_name(name)
        existing_roles_in_profile = self.info_for_profile(role_name)
        if existing_roles_in_profile is None:
            try:
                with self.catch_boto_400("Couldn't create instance profile", instance_profile=role_name):
                    for _ in self.change("+", "instance_profile", profile=role_name):
                        self.connection.create_instance_profile(role_name)
            except boto.exception.BotoServerError as error:
                if error.status == 409 and error.code == "EntityAlreadyExists":
                    # I'd rather ignore this conflict, than list all the instance_profiles
                    # Basically, the instance exists but isn't associated with the role
                    pass
                else:
                    raise

        if existing_roles_in_profile and any(rl != role_name for rl in existing_roles_in_profile):
            for role in [rl for rl in existing_roles_in_profile if rl != role_name]:
                with self.catch_boto_400("Couldn't remove role from an instance profile", profile=role_name, role=role):
                    for _ in self.change("-", "instance_profile_role", profile=role_name, role=role):
                        self.connection.remove_role_from_instance_profile(role_name, role)

        if not existing_roles_in_profile or not any(rl == role_name for rl in existing_roles_in_profile):
            with self.catch_boto_400("Couldn't add role to an instance profile", role=role_name, instance_profile=role_name):
                for _ in self.change("+", "instance_profile_role", profile=role_name, role=role_name):
                    self.connection.add_role_to_instance_profile(role_name, role_name)

    def create_role(self, name, trust_document, policies=None):
        """Create a role"""
        role_name, role_path = self.split_role_name(name)
        with self.catch_boto_400("Couldn't create role", "{0} trust document".format(name), trust_document, role=name):
            for _ in self.change("+", "role", role=role_name, document=trust_document):
                self.connection.create_role(role_name, assume_role_policy_document=trust_document, path=role_path)

        # And add our permissions
        if policies:
            for policy_name, document in policies.items():
                if document:
                    with self.catch_boto_400("Couldn't add policy", "{0} - {1} Permission document".format(role_name, policy_name), document, role=role_name, policy_name=policy_name):
                        for _ in self.change("+", "role_policy", role=role_name, policy=policy_name, document=document):
                            self.connection.put_role_policy(role_name, policy_name, document)

    def modify_role(self, role_info, name, trust_document, policies=LeaveAlone):
        """Modify a role"""
        role_name, _ = self.split_role_name(name)
        if trust_document:
            changes = list(self.documents.compare_trust_document(role_info, trust_document))
            if changes:
                with self.catch_boto_400("Couldn't modify trust document", "{0} assume document".format(role_name), trust_document, role=role_name):
                    for _ in self.change("M", "trust_document", role=role_name, changes=changes):
                        self.connection.update_assume_role_policy(role_name, trust_document)

        if policies is LeaveAlone:
            return
        elif policies is None:
            policies = {}

        unknown = []
        with self.catch_boto_400("Couldn't get policies for a role", role=role_name):
            current_policies = self.current_role_policies(role_name, comparing=[pn for pn in policies])
        unknown = [key for key in current_policies if key not in policies]

        if unknown:
            log.info("Role has unknown policies that will be disassociated\trole=%s\tunknown=%s", role_name, unknown)
            for policy in unknown:
                with self.catch_boto_400("Couldn't delete a policy from a role", policy=policy, role=role_name):
                    for _ in self.change("-", "role_policy", role=role_name, policy=policy):
                        self.connection.delete_role_policy(role_name, policy)

        for policy, document in policies.items():
            if not document:
                if policy in current_policies:
                    with self.catch_boto_400("Couldn't delete a policy from a role", policy=policy, role=role_name):
                        for _ in self.change("-", "policy", role=role_name, policy=policy):
                            self.connection.delete_role_policy(role_name, policy)
            else:
                needed = False
                changes = None

                if policy in current_policies:
                    changes = list(self.documents.compare_two_documents(current_policies.get(policy), document))
                    if changes:
                        log.info("Overriding existing policy\trole=%s\tpolicy=%s", role_name, policy)
                        needed = True
                else:
                    log.info("Adding policy to existing role\trole=%s\tpolicy=%s", role_name, policy)
                    needed = True

                if needed:
                    with self.catch_boto_400("Couldn't add policy document", "{0} - {1} policy document".format(role_name, policy), document, role=role_name, policy=policy):
                        symbol = "M" if changes else "+"
                        for _ in self.change(symbol, "role_policy", role=role_name, policy=policy, changes=changes, document=document):
                            self.connection.put_role_policy(role_name, policy, document)

    def current_role_policies(self, name, comparing):
        """Get the current policies for some role"""
        role_name, _ = self.split_role_name(name)
        with self.catch_boto_400("Couldn't get policies for a role", role=name):
            policies = self.connection.list_role_policies(role_name)["list_role_policies_response"]["list_role_policies_result"]["policy_names"]

        found = {}
        for policy in policies:
            document = None
            if policy in comparing:
                with self.catch_boto_400("Couldn't get policy document for some policy", policy=policy, role=name):
                    doc = self.connection.get_role_policy(role_name, policy)["get_role_policy_response"]["get_role_policy_result"]["policy_document"]
                document = json.dumps(json.loads(parse.unquote(doc)), indent=2).strip()
            found[policy] = document

        return found

    def remove_role(self, name):
        """Remove the role if it exists"""
        role_name, _ = self.split_role_name(name)
        if self.has_role(role_name):
            with self.catch_boto_400("Couldn't get policies for a role", role=role_name):
                current_policies = self.current_role_policies(role_name, comparing=[])

            for policy in current_policies:
                with self.catch_boto_400("Couldn't delete a policy from a role", policy=policy, role=role_name):
                    for _ in self.change("-", "policy", role=role_name, policy=policy):
                        self.connection.delete_role_policy(role_name, policy)

            with self.catch_boto_400("Couldn't delete a role", role=role_name):
                for _ in self.change("-", "role", role=role_name):
                    self.connection.delete_role(role_name)
        else:
            log.info("Role already deleted\trole=%s", role_name)




from iam_syncr.errors import BadAmazon

from contextlib import contextmanager
import boto

class LeaveAlone(object):
    """Used to differentiate between None and not specified in a call signature"""

class AmazonMixin:
    @contextmanager
    def catch_boto_400(self, message, heading=None, document=None, **info):
        """Turn a BotoServerError 400 into a BadAmazon"""
        try:
            yield
        except boto.exception.BotoServerError as error:
            if error.status == 400:
                if heading or document:
                    print("=" * 80)
                    if heading:
                        print(heading)
                    print(document)
                    print("=" * 80)
                raise BadAmazon(message, error_code=error.code, error_message=error.message, **info)
            else:
                raise

    @contextmanager
    def ignore_boto_404(self):
        """Ignore a BotoServerError 404"""
        try:
            yield
        except boto.exception.BotoServerError as error:
            if error.status == 404:
                pass
            else:
                raise

    def print_change(self, symbol, typ, changes=None, document=None, **kwargs):
        """Print out a change"""
        values = ", ".join("{0}={1}".format(key, val) for key, val in sorted(kwargs.items()))
        print("{0} {1}({2})".format(symbol, typ, values))
        if changes:
            for change in changes:
                print("\n".join("\t{0}".format(line) for line in change.split('\n')))
        elif document:
            print("\n".join("\t{0}".format(line) for line in document.split('\n')))

    def change(self, symbol, typ, **kwargs):
        """Print out a change and then do the change if not doing a dry run"""
        self.print_change(symbol, typ, **kwargs)
        if not self.amazon.dry_run:
            try:
                yield
            except:
                raise
            else:
                self.amazon.changes = True




from delfick_error import DelfickError, ProgrammerError

# I use Programmererror, make vim be quiet about it
ProgrammerError = ProgrammerError

class SyncrError(DelfickError): pass

class InvalidDocument(SyncrError):
    desc = "Something wrong with this iam document"

class BadConfiguration(SyncrError):
    desc = "Something wrong with the configuration"

class InvalidConfiguration(SyncrError):
    desc = "Something wrong with this configuration"

class NoConfiguration(SyncrError):
    desc = "Didn't find any yaml files"

class ConflictingConfiguration(SyncrError):
    desc = "The configuration conflicts with itself"

class DuplicateItem(SyncrError):
    desc = "Item defined multiple times"

class BadRole(SyncrError):
    desc = "Bad definition of a role"

class BadPolicy(SyncrError):
    desc = "Bad definition of a policy"

class BadAmazon(SyncrError):
    desc = "Amazon said no"

class CantFindTemplate(SyncrError):
    desc = "Can't find a template"

class NoTemplates(SyncrError):
    desc = "No templates defined"

class BadAlias(SyncrError):
    desc = "Bad kms alias"




def listify(dct, key):
    """Make sure a key in this dct is a list"""
    if key not in dct:
        dct[key] = []
    elif not isinstance(dct[key], list):
        dct[key] = [dct[key]]
    return dct[key]

def listified(dct, key):
    """
    Yield the items under key

    If item is not a list, then just yield that item
    """
    if key in dct:
        if isinstance(dct[key], list):
            for thing in dct[key]:
                yield thing
        else:
            yield dct[key]

def as_list(item):
    """Yields the items in the item unless not a list, in which case just yield the item"""
    if isinstance(item, list):
        for thing in item:
            yield thing
    else:
        yield item




from six.moves.urllib import parse
from datadiff import diff
import json

class AmazonDocuments(object):
    def compare_trust_document(self, role_info, trust_document):
        """Say whether the provided trust document is the same as the one in the role_info"""
        if not role_info or not role_info.get("role", {}).get("assume_role_policy_document"):
            return []

        unquoted = parse.unquote(role_info["role"]["assume_role_policy_document"])
        return self.compare_two_documents(unquoted, trust_document)

    def compare_two_documents(self, doc1, doc2):
        """Compare two documents by converting them into json objects and back to strings and compare"""
        try:
            first = json.loads(doc1)
        except (ValueError, TypeError):
            return

        try:
            second = json.loads(doc2)
        except (ValueError, TypeError):
            return

        # Ordering the principals because the ordering amazon gives me hates me
        def sort_statement(statement):
            for principal in (statement.get("Principal", None), statement.get("NotPrincipal", None)):
                if principal:
                    for principal_type in ("AWS", "Federated", "Service"):
                        if principal_type in principal and type(principal[principal_type]) is list:
                            principal[principal_type] = sorted(principal[principal_type])
        def sort_key(statement, key):
            if key in statement and type(statement[key]) is list:
                statement[key] = sorted(statement[key])
        for document in (first, second):
            if "Statement" in document:
                if type(document["Statement"]) is dict:
                    sort_statement(document["Statement"])
                    sort_key(document["Statement"], "Action")
                    sort_key(document["Statement"], "NotAction")
                    sort_key(document["Statement"], "Resource")
                    sort_key(document["Statement"], "NotResource")
                else:
                    for statement in document["Statement"]:
                        sort_statement(statement)
                        sort_key(statement, "Action")
                        sort_key(statement, "NotAction")
                        sort_key(statement, "Resource")
                        sort_key(statement, "NotResource")

        difference = diff(first, second, fromfile="current", tofile="new").stringify()
        if difference:
            lines = difference.split('\n')
            if not first or not second or any(line.strip().startswith("@@") and line.strip().endswith("@@") for line in lines):
                for line in lines:
                    yield line




from iam_syncr.amazon.buckets import AmazonBuckets
from iam_syncr.statements import Statements
from iam_syncr.helpers import listified
from iam_syncr.errors import SyncrError

import logging
import six

log = logging.getLogger("iam_syncr.buckets")

class Bucket(object):
    def __init__(self, name, definition, amazon, templates=None):
        self.name = name
        self.definition = definition
        self.statements = Statements(name, "bucket", amazon.account_id, amazon.accounts)

        self.amazon = amazon
        self.amazon_buckets = AmazonBuckets(amazon)

        self.permission = []

    def setup(self):
        """Raise errors if the definition doesn't make sense"""
        self.tags = self.definition.get("tags", {})
        if type(self.tags) != dict or any(not isinstance(tag_name, six.string_types) or not isinstance(tag_val, six.string_types) for tag_name, tag_val in self.tags.items()):
            raise SyncrError("Bucket tags should be a dictionary of {<string> : <string>}", got=self.tags)

        self.location = self.definition.get("location", "ap-southeast-2")
        for key, default_allow in (("permission", None), ("allow_permission", True), ("deny_permission", False)):
            for policy in listified(self.definition, key):
                for statement in self.statements.make_permission_statements(policy, allow=default_allow):
                    self.permission.append(statement)

    def resolve(self):
        """Make sure this user exists and has only what policies we want it to have"""
        permission_document = self.make_permission_document(self.permission)

        bucket_info = self.amazon_buckets.bucket_info(self.name)
        if not bucket_info:
            self.amazon_buckets.create_bucket(self.name, self.location, permission_document=permission_document, tags=self.tags)
        else:
            self.amazon_buckets.modify_bucket(self.name, self.location, permission_document=permission_document, tags=self.tags)

    def make_permission_document(self, permissions):
        """Return a document for these permissions, or None if no permissiosn"""
        if not permissions:
            return
        return self.statements.make_document(permissions)

